import PrepRound from "./prepRound";
import Package from "./package";

export default class Prep {
  public fieldsToKeep: string[] = [];
  public prepRounds: PrepRound[] = [];
  public manPrepPackages: Package[] = [];
}
